package com.wmstein.tourcount

import android.annotation.SuppressLint
import android.content.Intent
import android.graphics.Bitmap
import android.graphics.drawable.BitmapDrawable
import android.os.Bundle
import android.util.Log
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.WindowManager
import android.widget.LinearLayout
import android.widget.ScrollView
import androidx.appcompat.app.AppCompatActivity
import androidx.core.app.NavUtils
import com.wmstein.tourcount.database.Count
import com.wmstein.tourcount.database.CountDataSource
import com.wmstein.tourcount.widgets.SpeciesAddWidget

/************************************************************************
 * AddSpeciesActivity lets you insert a new species into the species list
 * AddSpeciesActivity is called from EditSpecListActivity
 * Uses SpeciesAddWidget.kt, widget_add_spec.xml.
 *
 * The sorting order of the species to add cannot be changed, as it is determined
 * by 3 interdependent and correlated arrays in arrays.xml
 *
 * Created for TourCount by wmstein on 2019-04-12,
 * last edited in Java on 2023-05-13,
 * converted to Kotlin on 2023-07-06
 * last edited on 2024-05-15
 */
class AddSpeciesActivity : AppCompatActivity() {
    private var tourCount: TourCountApplication? = null

    private var addArea: LinearLayout? = null

    // the actual count data
    private var countDataSource: CountDataSource? = null

    // Id list of missing species
    private lateinit var idArray: Array<String?>

    // complete ArrayLists of species
    private var namesCompleteArrayList: ArrayList<String>? = null
    private var namesGCompleteArrayList: ArrayList<String>? = null
    private var codesCompleteArrayList: ArrayList<String?>? = null
    private var specName: String? = null
    private var specCode: String? = null
    private var specNameG: String? = null // selected species
    private var bMap: Bitmap? = null
    private var bg: BitmapDrawable? = null

    // preferences
    private var prefs = TourCountApplication.getPrefs()
    private var brightPref = false

    @SuppressLint("SourceLockedOrientationActivity")
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        tourCount = application as TourCountApplication

        // Load preferences
        brightPref = prefs.getBoolean("pref_bright", true)

        if (MyDebug.LOG) Log.d(TAG, "68, onCreate")

        setContentView(R.layout.activity_add_species)
        val addScreen = findViewById<ScrollView>(R.id.add_screen)

        // Set full brightness of screen
        if (brightPref) {
            window.addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON)
            val params = window.attributes
            params.screenBrightness = 1.0f
            window.attributes = params
        }

        bMap = tourCount!!.decodeBitmap(
            R.drawable.abackground,
            tourCount!!.width,
            tourCount!!.height
        )
        bg = BitmapDrawable(addScreen.resources, bMap)
        addScreen.background = bg

        addArea = findViewById(R.id.addSpecLayout)

        // Load complete species ArrayList from arrays.xml (lists are sorted by code)
        namesCompleteArrayList = ArrayList(listOf(*resources.getStringArray(R.array.selSpecs)))
        namesGCompleteArrayList = ArrayList(listOf(*resources.getStringArray(R.array.selSpecs_g)))
        codesCompleteArrayList = ArrayList(listOf(*resources.getStringArray(R.array.selCodes)))
    }
    // end of onCreate()

    override fun onResume() {
        super.onResume()

        // clear any existing views
        addArea!!.removeAllViews()

        // setup the data sources
        countDataSource = CountDataSource(this)
        countDataSource!!.open()

        supportActionBar!!.setTitle(R.string.addTitle)

        // list only new species not already contained in the species counting list

        // 1. code list of contained species
        val specCodesContainedList = ArrayList<String?>()

        // get species of the counting list
        val counts: List<Count> = countDataSource!!.allSpeciesSrtCode

        // build code ArrayList of already contained species
        for (count in counts) {
            specCodesContainedList.add(count.code)
        }

        // 2. build lists of missing species
        val specCodesContainedListSize = specCodesContainedList.size
        var posSpec: Int

        // for already contained species reduce complete arraylists
        for (i in 0 until specCodesContainedListSize) {
            if (codesCompleteArrayList!!.contains(specCodesContainedList[i])) {
                // Remove species with code x from missing species lists.
                // Prerequisites: exactly correlated arrays of selCodes, selSpecs and selSpecs_g
                //   for all localisations
                specCode = specCodesContainedList[i]
                posSpec = codesCompleteArrayList!!.indexOf(specCode)
                namesCompleteArrayList!!.removeAt(posSpec)
                namesGCompleteArrayList!!.removeAt(posSpec)
                codesCompleteArrayList!!.remove(specCode)
            }
        }
        idArray = setIdsSelSpecs(codesCompleteArrayList) // create idArray from codeArray

        // load the species data into the widgets
        var i = 0
        while (i < codesCompleteArrayList!!.size) {
            val saw = SpeciesAddWidget(this, null)
            saw.setSpecName(namesCompleteArrayList!![i])
            saw.setSpecNameG(namesGCompleteArrayList!![i])
            saw.setSpecCode(codesCompleteArrayList!![i])
            saw.setPSpec(codesCompleteArrayList!![i]!!)
            saw.setSpecId(idArray[i]!!)
            addArea!!.addView(saw)
            i++
        }
    } // end of Resume()

    // create idArray from codeArray
    private fun setIdsSelSpecs(speccodesm: ArrayList<String?>?): Array<String?> {
        idArray = arrayOfNulls(speccodesm!!.size)
        var i = 0
        while (i < speccodesm.size) {
            idArray[i] = (i + 1).toString()
            i++
        }
        return idArray
    }

    override fun onSaveInstanceState(outState: Bundle) {
        outState.putString("new_spec_code", specCode)
        super.onSaveInstanceState(outState)
    }

    public override fun onRestoreInstanceState(savedInstanceState: Bundle) {
        if (savedInstanceState.getString("new_spec_code")!!.isNotBlank())
            specCode = savedInstanceState.getString("new_spec_code")
        super.onRestoreInstanceState(savedInstanceState)
    }

    override fun onPause() {
        super.onPause()

        countDataSource!!.close()
    }

    fun saveAndExit(view: View) {
        if (saveData(view)) {
            super.finish()
        }
    }

    private fun saveData(view: View): Boolean {
        // save added species to species list
        var retValue = true
        val idToAdd = view.tag as Int

        val saw1 = addArea!!.getChildAt(idToAdd) as SpeciesAddWidget
        specName = saw1.getSpecName()
        specCode = saw1.getSpecCode()
        specNameG = saw1.getSpecNameG()

        try {
            countDataSource!!.createCount(specName, specCode, specNameG)
        } catch (e: Exception) {
            retValue = false
        }

        // store code of new species in sharedPreferences for Spinner in CountingActivity(A)
        val editor = prefs.edit()
        editor.putString("new_spec_code", specCode)
        editor.commit()

        return retValue
    }

    // Add the selected species to the species list
    fun addCount(view: View) {
        if (saveData(view)) {
            super.finish()
        }
    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        // Inflate the menu; this adds items to the action bar if it is present.
        menuInflater.inflate(R.menu.add_species, menu)
        return true
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        // Handle action bar item clicks here. The action bar will
        // automatically handle clicks on the Home/Up button, so long
        // as you specify a parent activity in AndroidManifest.xml.
        val id = item.itemId
        if (id == android.R.id.home) {
            val intent = NavUtils.getParentActivityIntent(this)!!
            intent.flags = Intent.FLAG_ACTIVITY_CLEAR_TOP or Intent.FLAG_ACTIVITY_SINGLE_TOP
            NavUtils.navigateUpTo(this, intent)
        }
        return super.onOptionsItemSelected(item)
    }

    @Deprecated("Deprecated in Java")
    @SuppressLint("ApplySharedPref", "MissingSuperCall")
    override fun onBackPressed() {
        countDataSource!!.close()
        finish()
//        NavUtils.navigateUpFromSameTask(this)
//        @Suppress("DEPRECATION")
//        super.onBackPressed()
    }

    companion object {
        private const val TAG = "AddSpecAct"
    }

}
